#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#include <vgagames3.h>
#include "main.h"
#include "obj-asteroid.h"

void getofc_asteroid(struct vg3_ofunc_objfunc *);

static struct vg3_ofunc_object * f_new(void *, unsigned int, ...);
static void f_free(void *, struct vg3_ofunc_object *);
static void f_run(void *, struct vg3_ofunc_object *);
static void f_draw(void *, struct vg3_ofunc_object *);


/* +++ get-function +++ */

void
getofc_asteroid(struct vg3_ofunc_objfunc *ofc)
{
  if (ofc == NULL) { return; }

  snprintf(ofc->oid, sizeof(ofc->oid), "%s", get_oid_name(OID_NAME_ASTEROID));
  ofc->f_new = f_new;
  ofc->f_free = f_free;
  ofc->f_run = f_run;
  ofc->f_draw = f_draw;
}


/* +++ object-instance-functions +++ */

/* new-function
 * variable parameters:
 * - int: x-position (centered)
 * - int: y-position (centered)
 * - int: x-moving
 * - int: y-moving
 * - int: subID
 * - int: which sprite
 */
static struct vg3_ofunc_object *
f_new(void *vmain, unsigned int iparent, ...)
{
  struct g_main *gmain = vmain;
  struct vg3_ofunc_object *objp;
  struct g_obj_asteroid *gobj;
  struct vg3_coll coll;
  va_list ap;
  int xm, ym, xd, yd, subid, spriteid;
  char buf[128];

  if (gmain == NULL) { VG3_seterror(EINVAL, strerror(EINVAL)); return NULL; }

  /* get arguments */
  va_start(ap, iparent);
  xm = va_arg(ap, int);
  ym = va_arg(ap, int);
  xd = va_arg(ap, int);
  yd = va_arg(ap, int);
  subid = va_arg(ap, int);
  spriteid = va_arg(ap, int);
  va_end(ap);

  /* create private struct for object */
  gobj = calloc(1, sizeof(*gobj));
  if (gobj == NULL) { VG3_seterror(ENOMEM, strerror(errno)); return NULL; }

  /* load sprites */
  snprintf(buf, sizeof(buf), "%s/bmp/asteroid/asteroid%d-%d.sprite", FILES_DIR, subid, spriteid);
  gobj->images.sprt = VG3_sprite_load(gmain->wstruct, buf);
  if (gobj->images.sprt == NULL) { return NULL; }
  gobj->images.hit = VG3_sprite_load(gmain->wstruct, FILES_DIR "/bmp/hit/hit.sprite");
  if (gobj->images.hit == NULL) { return NULL; }
  gobj->images.expl = VG3_sprite_load(gmain->wstruct, FILES_DIR "/bmp/expl/expl1.sprite");
  if (gobj->images.expl == NULL) { return NULL; }

  /* set position */
  VG3_sprite_imagesize(gobj->images.sprt, "avg", 100, &gobj->rect.w, &gobj->rect.h);
  gobj->rect.x = xm - gobj->rect.w / 2;
  gobj->rect.y = ym - gobj->rect.h / 2;
  /* set collision rectangle with 60% of width and height of image-rectangle */
  gobj->crect = gobj->rect;
  gobj->crect = VG3_correct_imageposition(&gobj->crect, gobj->rect.w * 60 / 100, gobj->rect.h * 60 / 100);

  /* set speed */
  gobj->xdelta = xd;
  gobj->ydelta = yd;

  /* set health */
  gobj->max_health = gobj->health = (subid == 1 ? 1 : 2);

  /* create and fill asteroid-instance */
  objp = calloc(1, sizeof(*objp));
  if (objp == NULL) { VG3_seterror(ENOMEM, strerror(errno)); return NULL; }
  snprintf(objp->oid, sizeof(objp->oid), "%s", get_oid_name(OID_NAME_ASTEROID));
  objp->subid = subid;
  objp->drawlevel = 2;
  objp->instanceid = 0;  /* will be set in VG3_ofunc_objlist_insert() */
  objp->ostruct = gobj;

  /* insert asteroid-instance into list of object-instances */
  VG3_ofunc_objlist_insert(gmain->ofstruct, objp);

  /* insert asteroid-instance (collision-rectangle) into quadtree */
  memset(&coll, 0, sizeof(coll));
  coll.rect = gobj->crect;
  snprintf(coll.oid, sizeof(coll.oid), "%s", objp->oid);
  coll.optr = objp;
  VG3_coll_q_insert(gmain->qdtr, &coll);

  return objp;
}


/* free-function */
static void
f_free(void *vmain, struct vg3_ofunc_object *objp)
{
  struct g_main *gmain = vmain;
  struct g_obj_asteroid *gobj;

  if (gmain == NULL || objp == NULL) { return; }

  gobj = (struct g_obj_asteroid *)objp->ostruct;

  /* remove asteroid-instance from quadtree */
  VG3_coll_q_remove(gmain->qdtr, objp);

  /* remove asteroid-instance from list of object-instances */
  VG3_ofunc_objlist_remove(gmain->ofstruct, objp);

  /* free asteroid-instance */
  if (gobj->images.sprt != NULL) { VG3_sprite_free(gobj->images.sprt); }
  if (gobj->images.hit != NULL) { VG3_sprite_free(gobj->images.hit); }
  if (gobj->images.expl != NULL) { VG3_sprite_free(gobj->images.expl); }
  free(gobj);
  free(objp);
}


/* run-function */
static void
f_run(void *vmain, struct vg3_ofunc_object *objp)
{
  const int moving_factor = 10;
  struct g_main *gmain = vmain;
  struct g_obj_asteroid *gobj;
  struct vg3_coll coll;
  int erg;

  if (gmain == NULL || objp == NULL) { return; }

  gobj = (struct g_obj_asteroid *)objp->ostruct;

  /* remove asteroid-instance from quadtree */
  VG3_coll_q_remove(gmain->qdtr, objp);

  /* check if asteroid is exploding */
  if (gobj->health <= 0) { return; }

  /* check if asteroid is gone */
  if (gobj->xdelta < 0 && gobj->rect.x < -gobj->rect.w) { f_free(gmain, objp); return; }
  if (gobj->xdelta > 0 && gobj->rect.x > gmain->winw) { f_free(gmain, objp); return; }
  if (gobj->ydelta < 0 && gobj->rect.y < -gobj->rect.h) { f_free(gmain, objp); return; }
  if (gobj->ydelta > 0 && gobj->rect.y > gmain->winh) { f_free(gmain, objp); return; }

  /* move asteroid-instance and check for collisions with gobj->crect */
  erg = VG3_move_object_check_collision(gmain, gmain->ofstruct, gmain->qdtr, objp, &gobj->crect, moving_factor, &gobj->xdelta, &gobj->ydelta, &gobj->xremainder, &gobj->yremainder);
  if (erg < 0) {  /* error: destroy asteroid-instance */
    fprintf(stderr, "moving asteroid: %s\n", VG3_error());
    f_free(gmain, objp);
    return;
  }
  if (erg == VGAG3_COLL_RETURN_DEAD) { return; }  /* instance has been freed */
  if (erg == VGAG3_COLL_RETURN_CATCHED) { f_free(gmain, objp); return; }

  /* x/y-value of gobj->crect could be modified, correct it also in gobj->rect */
  gobj->rect = VG3_correct_imageposition(&gobj->crect, gobj->rect.w, gobj->rect.h);

  /* insert asteroid-instance (collision-rectangle) into quadtree again */
  memset(&coll, 0, sizeof(coll));
  coll.rect = gobj->crect;
  snprintf(coll.oid, sizeof(coll.oid), "%s", objp->oid);
  coll.optr = objp;
  VG3_coll_q_insert(gmain->qdtr, &coll);
}


/* draw-function */
static void
f_draw(void *vmain, struct vg3_ofunc_object *objp)
{
  struct g_main *gmain = vmain;
  struct g_obj_asteroid *gobj;
  struct vg3_image *img;
  struct vg3_image_attributes iattr;

  if (gmain == NULL || objp == NULL) { return; }

  gobj = (struct g_obj_asteroid *)objp->ostruct;

  /* draw asteroid-instance */
  if (gobj->health <= 0) {  /* exploding */
    if (!VG3_sprite_get(gobj->images.expl, &img, &iattr, NULL)) { f_free(gmain, objp); return; }
  } else {
    if (!VG3_sprite_get(gobj->images.sprt, &img, &iattr, NULL)) {
      VG3_sprite_rewind(gobj->images.sprt);
      VG3_sprite_get(gobj->images.sprt, &img, &iattr, NULL);
    }
  }
  if (img != NULL) {
    VG3_image_copy(gmain->wstruct, NULL, img, gobj->rect.x + gobj->rect.w / 2, gobj->rect.y + gobj->rect.h / 2, &iattr, 0);
    if (gobj->health > 0 && gobj->hit_y > 0) {  /* being hit */
      if (!VG3_sprite_get(gobj->images.hit, &img, &iattr, NULL)) {
        VG3_sprite_rewind(gobj->images.hit);
        gobj->hit_y = 0;
      } else if (img != NULL) {
        VG3_image_copy(gmain->wstruct, NULL, img, gobj->rect.x + gobj->rect.w / 2, gobj->rect.y + gobj->hit_y, &iattr, 0);
        draw_health(vmain, gobj->max_health, gobj->health, 0);
      }
    }
  }
}
